#!/usr/bin/perl -w

# Hopfield network for letter recognition

readpatterns();
flush();
fixpatterns();
flush();
init_hop(5);
foreach $i (0 .. 25)
	{
	my @topr = runon($i);
	pgrid(@topr);
	}


sub pgrid
			# Prints the results of a state of the nwk
{
my @adat = @_;
my $ycount = 0;
while($ycount < 15)
	{
	my $xcount = 0;
	while($xcount < 15)
		{
		my $coord = 0;
		my $zindex = di($xcount, $ycount);
		$coord = $adat[$zindex];
		if($coord == -1)
			{
			print "0 ";
			}
		else
			{
			print "1 ";
			}
		$xcount++;
		}
	print "\n";
	$ycount++;
	}
print "\n\n";
}


sub assignpat
{
my $toas = shift(@_);	# Patternid to assign into the network
my $tptr;
$tptr = $patterns[$toas];
@nwkstate = @$tptr;
}


sub runonce
			# Runs hopfield network on a single value update
{			# Operates on @nwkstate
my @nextstate = ();	# Temporary network state
my $walker = 0;		# Where we are
while($walker < 225)
	{
	my $runner = 0;	# Which input we're thinking about
	$nextstate[$walker] = 0;
	while($runner < 225)
		{
		$nextstate[$walker] += $weights[$walker][$runner] * $nwkstate[$runner];
		$runner++;
		}
			# EXPER
	if($nextstate[$walker] > 0)
		{
		$nextstate[$walker] = 1;
		}
	if($nextstate[$walker] < 0)
		{
		$nextstate[$walker] = -1;
		}
			# E EXPER
	$walker++;
	}
return @nextstate;
}


sub runon
			# Runs hopfield network on a given patternid
			# Returns array of final result
{
my $patternid = shift(@_);
assignpat($patternid);		# Inputs into the network
print "Running hopfield network on pattern $patternid";
@nwkstate = runonce();	# The chicken before the egg
@onwkstate = @nwkstate;

ploop: while(1)
	{
	print ".";
	flush();
	@nwkstate = runonce();
		# See if @nresults logically equal @results
		# if so, stop. Else, continue
	my $crawler = 0;
	while($crawler < 225)
		{
		if($nwkstate[$crawler] != $onwkstate[$crawler])
			{
#			print "[$crawler:" . $nwkstate[$crawler] . ":" . $onwkstate[$crawler] . "]";
			@onwkstate = @nwkstate; # The new shall become old
			next ploop;	# All new world
			}
		$crawler++;
		}
					# At this point, they're the same!
					# Huzzah!
	print "\n";
	return @nwkstate;
	}
}


sub init_hop
			# Initialize hopfield network
{
my $numpat = shift(@_);	# How many patterns we want to init with
die "Can't init with zero input" if($numpat == 0);
print "Initializing hopfield network with $numpat patterns";
my $ptra = 0;		# First pointer
my $ptrb = 0;		# Second pointer
while($ptra < 225)
	{
	$ptrb = 0;
	flush();
	print ".";
	iloop: while($ptrb < 225)
		{
		my $patptr = 0;
		if($ptra == $ptrb)
			{
			$weights[$ptra][$ptrb] = 0;
			$ptrb++;
			next iloop;
			}
		$weights[$ptra][$ptrb] = 0; # I could optimize here, but it'd
					    # cause a loss of clarity
		while($patptr < $numpat)
			{
			$weights[$ptra][$ptrb] += ($patterns[$patptr][$ptra] * $patterns[$patptr][$ptrb]);
			$patptr++;
			}
		$ptrb++;
		}
	$ptra++;
	}
print "\n";
}


sub fixpatterns
			# The patterns as read in are in binary 1/0 format.
			# We must correct them to 1/-1 here
{
print "Correcting patterns";
my $pctr = 0; 		# Pattern Counter
my $valctr = 0;		# Value counter
while($pctr < 26)
	{
	flush();
	print ".";
	$valctr = 0;
	while($valctr < 225)
		{
		if( $patterns[$pctr][$valctr] == 0 )
			{
			$patterns[$pctr][$valctr] = -1;
			}
		$valctr++;
		}
	$pctr++;
	}
print "\n";
}


sub readpatterns
			# Reads in all 26 letters in from file
{
my $pcount = 0; # Counter for what pattern we're on
open(PDATA, "proj-3.data") || die "Cannot open input patterns file proj-3.data";
while($pcount < 26)	# Remember, we're zero-based
	{
	print "Reading pattern $pcount\n";
	my @thisarr = ();
	my $clcount = 0; # What line of the current pattern we're on
	while($clcount < 15)
		# Read a line, pack it into the appropriate structures
		{
		my $tline = <PDATA>;
		chomp $tline;
		my @lar = split(/ /, $tline);
#		print "Data " . "@lar" . "\n";
		push(@thisarr, @lar);
		$clcount++;
		}
	$patterns[$pcount] = \@thisarr;
#	print "\$patterns\[$pcount\] = @thisarr \n";
#	print "Slimy " . $patterns[$pcount][0] . "\n"; 
	$tline = <PDATA>;
	chomp($tline);
#	print "Blank? " . $tline . "\n";
	if(! ($tline =~ /^\s*$/) ) {die "Bad format!"}
#	print "Slick " . $patterns[$pcount][0] . "\n";
	$pcount++;
	}
}




sub di
			# Translate 2d indices into 1d index
{
my $xi = shift(@_);
my $yi = shift(@_);
if(($yi > 14) || ($xi > 14) || ($xi < 0) || ($yi < 0))
	{
	die "Invalid hopnet acc ($xi, $yi)";
	}
return ($xi + (15*$yi));
}

sub flush
			# Not very readable, I know. Flushes the
			# output buffer.
{
my $ta = 1;
($|, $ta) = ($ta, $|);
($|, $ta) = ($ta, $|);
}
